#!/usr/bin/env python

"""Helper class for printing

This module contains helper functions to beautify or ease
the printing process to standard and error output.

It contains a color and a markup class, defining specific ANSI codes.
Moreover, four status badges for console output are defined,
as well as two wrapper functions around click.echo.

    * Color - color class containing ANSI color codes.
    * Markup - markup class containing ANSI markup codes.
    * echov - Wrapping around click.echo and adding a verbose option.
    * echof - Print to error output and exit afterwards.
"""
import os
import sys

import click

if sys.platform.lower() == "win32":
    os.system("color")


class _Color:
    """ A simple color class containing ANSI color codes. """

    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"


class _Markup:
    """ A simple markup class containing ANSI color codes. """

    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


_INFO = _Markup.BOLD + _Color.PURPLE + "[STATUS] " + _Markup.END
_USAGE = _Markup.BOLD + _Color.YELLOW + "[USAGE] " + _Markup.END
_FAIL = _Markup.BOLD + _Color.RED + "[ERROR] " + _Markup.END
_WARNING = _Markup.BOLD + _Color.YELLOW + "[WARNING] " + _Markup.END


def echov(text, verbose=True, new_line=True):
    """ Wraps around click.echo and expands it by a `verbose` option.

    Parameter
    ---------
    text: str
        Text that is printed to standard output.
    verbose: {True, False}, optional
        Determines whether something is printed at all.
    new_line: {True, False}, optional
        If set, a new line is added to the output.
    """
    if verbose:
        click.echo(_INFO + text, nl=new_line)


def echoe(text, error=None):
    """ Prints text to error output using click.echo and exits.

    Parameter
    ---------
    text: str
        Text that is printed to error output.
    """
    click.echo(_FAIL + text, err=True)
    if error:
        raise error
    sys.exit(1)


def echow(text):
    """ Prints text warning to standard output and prints message.

    Parameter
    ---------
    text: str
        Text that is printed to standard output.
    """
    click.echo(_WARNING + text)



#!/usr/bin/env python

"""Helper class for communicating with YouTube API v3

This module contains helper function to communicate with
Google"s YouTube API v3.
"""

from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from ratelimit import limits, sleep_and_retry

from ytscraper.helper.echo import echow, echoe


def related_search(handle, number, videoId, **api_options):
    api_options["relatedToVideoId"] = videoId
    api_options["maxResults"] = number + 1
    return _search(handle, number, **api_options)


def video_search(handle, number, search_term, **api_options):
    api_options["q"] = search_term
    api_options["maxResults"] = number
    return _search(handle, number, **api_options)


def _search(handle, number, **api_options):
    api_options["type"] = "video"
    api_options["part"] = "id,snippet"
    response = _get_response(handle.search(), **api_options)
    video_data = [_extract_data(video) for video in response["items"]]
    if len(video_data) != number:
        echow(
            f"API gave different number of videos than requested: {len(video_data)} instead of {number}!"
        )
    return video_data[:number]


def video_info(handle, video_id, **api_options):
    api_options["id"] = video_id
    api_options["part"] = "id,snippet"
    response = _get_response(handle.videos(), **api_options)
    video_data = [_extract_data(video) for video in response["items"]]
    return video_data


@sleep_and_retry
@limits(3_000_000, 100)
def _get_response(mod_handle, **api_options):
    try:
        return mod_handle.list(**api_options).execute()
    except HttpError as e:
        echoe("I am sorry to inform you that this program is going to fail!", e)


def _extract_data(item):
    result = {}
    if "id" in item:
        # Result between search term and relatedToVideoId differs
        try:
            result["videoId"] = item["id"]["videoId"]
        except TypeError:
            result["videoId"] = item["id"]
    if "snippet" in item:
        result["title"] = item["snippet"]["title"]
        result["description"] = item["snippet"]["description"]
        result["publishedAt"] = item["snippet"]["publishedAt"]
        result["channdelId"] = item["snippet"]["channelId"]
        result["channelTitle"] = item["snippet"]["channelTitle"]
    return result


def get_youtube_handle(api_key):
    """ Returns the YouTube Data API v3 handle.

    Parameter
    ---------
    api_key: str
        An authorized API key.

    Returns
    -------
        An object for interacting with the YouTube API v3 service.
    """
    try:
        return build("youtube", "v3", developerKey=api_key)
    except HttpError:
        echoe(
            """ There was an error while connecting to the YouTube API.
        Please check your API key."""
        )



#!/usr/bin/env python

from collections import deque
from pprint import pprint
from urllib import parse

import click

from ytscraper.helper.configfile import update_config
from ytscraper.helper.echo import echoe, echov
from ytscraper.helper.export import export_to_csv
from ytscraper.helper.yt_api import (
    get_youtube_handle,
    video_search,
    video_info,
    related_search,
)

verbose = False


@click.command()
@click.argument("search-type", default="term", type=click.Choice(["term", "url", "id"]))
@click.argument("query", nargs=1, required=True)
@click.option(
    "--number",
    "-n",
    multiple=True,
    type=click.IntRange(1, 50),
    help="Number of videos fetched per level.",
)
@click.option(
    "--max-depth",
    "-d",
    type=click.IntRange(0, 100),
    help="Maximal number of recursion steps.",
)
@click.option("--api-key", "-k", type=str, help="API Key to use YouTube API v3.")
@click.option(
    "--output-dir",
    "-o",
    type=click.Path(file_okay=False, writable=True),
    default="",
    help="Path to directory where output files are saved.",
)
@click.option(
    "--output-format",
    "-f",
    type=click.Choice(["csv"]),
    help="The file format of output files.",
)
@click.option(
    "--region-code",
    "-r",
    type=str,
    help="Return only videos that are unrestricted in a specified region.",
)
@click.option(
    "--lang-code",
    "-l",
    type=str,
    help="Return videos mostly relevant to a specified language.",
)
@click.option(
    "--safe-search",
    "-s",
    type=click.Choice(["none", "moderate", "strict"]),
    help="Return filtered results.",
)
@click.pass_context
def search(context, search_type, query, **options):
    """Searches YouTube using a specified query."""

    global verbose
    verbose = context.obj["verbose"]

    config = get_config(context, options)
    validate(config)

    rename = {
        "region_code": "regionCode",
        "lang_code": "relevanceLanguage",
        "safe_search": "safeSearch",
    }
    api_options = {rename[key]: config[key] for key in rename if config[key]}
    handle = get_handle(config["api_key"])

    start_videos = get_starter_videos(config, handle, api_options, search_type, query)
    nodes = build_nodes(config, handle, api_options, start_videos)

    if config["output_dir"] and config["output_format"] == "csv":
        echov("Query finished! Start exporting files!", verbose)
        export_to_csv(nodes, config["output_dir"])
        echov(f"Query finished! Exported results to {config['output_dir']}.")
    else:
        echov("Query finished! Result:")
        for node in nodes:
            print(
                "    " * node["depth"],
                f"Depth: {node['depth']}, Rank: {node['rank']}, ID: {node['videoId']}",
            )
            print("    " * node["depth"], f"           Title: {node['title']}")
            print(
                "    " * node["depth"],
                "           Related Videos: {}".format(node.get("relatedVideos")),
            )


def get_config(context, options):
    """ Reads the configuration file and updates it
    with the given command-line options. """
    config = context.obj["config"]
    echov("Updating configuration with command line options.", verbose)

    update_config(config, options)
    echov("Done! Working with the following configuration:", verbose)
    if verbose:
        pprint(config)
    return config


def validate(config):
    """ Checks validity of configuration. """
    config["number"] = tuple(config["number"],)
    config["max_depth"] = int(config["max_depth"])


def get_handle(api_key):
    """ Obtains the YouTube resource handle using an API key. """
    echov("Starting YouTube authentication.", verbose)
    if not api_key:
        echoe(
            """You need to provide an API key using `--api-key`
        or the configuration file in order to query YouTube's API.
        Please see README on how to obtain such a key."""
        )
    handle = get_youtube_handle(api_key)
    echov("API access established.", verbose)
    return handle


def get_starter_videos(config, handle, api_options, search_type, query):
    echov("Starting search using query {query}.", verbose)
    if search_type == "term":
        return video_search(handle, config["number"][0], query, **api_options)
    if search_type == "id":
        return video_info(handle, query)
    if search_type == "url":
        # Parse URL
        qterm = parse.urlsplit(query).query
        video_id = parse.parse_qs(qterm)["v"][0]
        return video_info(handle, video_id)
    raise click.BadParameter("Wrong search type.")


def build_nodes(config, handle, api_options, starter_videos):
    for rank, video in enumerate(starter_videos):
        video.update({"rank": rank, "depth": 0})
    queue = deque(starter_videos)
    processed = []
    while len(queue) > 0:
        video = queue.popleft()
        echov(
            f"Processing video {video['videoId']} (Depth: {video['depth']}).", verbose
        )
        processed.append(video)
        if video["depth"] >= config["max_depth"]:
            video["relatedVideos"] = list()
            continue
        # Add children
        num_children = _get_branching(config["number"], video["depth"])
        children = related_search(handle, num_children, video["videoId"], **api_options)
        video["relatedVideos"] = list(map(lambda c: c["videoId"], children))
        for rank, child in enumerate(children):
            child.update({"rank": rank, "depth": video["depth"] + 1})
        queue.extend(children)
    return processed


def _get_branching(container, index):
    """ Returns the nearest valid element from an interable container.

    This helper function returns an element from an iterable container.
    If the given `index` is not valid within the `container`,
    the function returns the closest element instead.

    Parameter
    ---------
        container:
            A non-empty iterable object.
        index:
            The index of an element that should be returned.

    Returns
    -------
        object
            The closest possible element from `container` for `index`.

    Example
    -------
    The `container` can be an arbitrary iterable object such as a list::

        l = ['a', 'b', 'c']
        c1 = _get_clamped_index(l, 5)
        c2 = _get_clamped_index(l, 1)
        c3 = _get_clamped_index(l, -4)

    The first call of the function using an index of 5 will return element 'c',
    the second call will return 'b' and the third call will return 'a'.
    """
    maximal_index = len(container) - 1
    minimal_index = 0
    clamped_index = min(maximal_index, index)
    clamped_index = max(minimal_index, clamped_index)
    return container[clamped_index]



#!/usr/bin/env python

import click

from ytscraper.helper.echo import echov, echow
from ytscraper.helper.configfile import load_config, write_config
from ytscraper.helper.configfile import DEFAULT_OPTIONS
from pprint import pprint
import ast


@click.group()
@click.pass_context
def config(context):
    """ Shows and modifies default configurations. """
    verbose = context.obj["verbose"]
    echov("Starting YTcrawl's config mode.", verbose)
    echov("Read the following configuration:", verbose)
    if verbose:
        pprint(context.obj["config"])
    pass


@config.command()
@click.argument("option", type=click.Choice(DEFAULT_OPTIONS), metavar="KEY")
@click.argument("value")
@click.pass_context
def set(context, option, value):
    """ Sets default options."""
    config = context.obj["config"]
    verbose = context.obj["verbose"]
    config_path = context.obj["config_path"]

    try:
        value = ast.literal_eval(value)
    except:
        if value == "true":
            value = True
        if value == "false":
            value = False
    # TODO does not support List or Tuple types
    target_type = type(DEFAULT_OPTIONS[option])
    if not isinstance(value, target_type):
        raise click.BadArgumentUsage(
            f"Given value '{value}' is not a valid type for {option}. Please provide a {target_type.__name__}."
        )
    elif target_type is int and value < 0:
        raise click.BadArgumentUsage(
            f"Given integer {value} is negative! Please provide a non-negative value.."
        )

    config[option] = value
    echov("The new configurations file is:", verbose)
    if verbose:
        pprint(config)
    write_config(config, config_path)
    echov("Successfully changed!")


@config.command()
@click.argument("option", type=click.Choice(DEFAULT_OPTIONS), metavar="KEY")
@click.pass_context
def unset(context, option):
    """ Unsets a default option."""
    config = context.obj["config"]
    verbose = context.obj["verbose"]
    config_path = context.obj["config_path"]

    if option in config:
        del config[option]

    echov("The new configurations file is:", verbose)
    if verbose:
        pprint(config)
    write_config(config, config_path)
    echov("Successfully written!")


@config.command()
@click.argument("option", type=click.Choice(DEFAULT_OPTIONS), metavar="KEY")
@click.pass_context
def get(context, option):
    """ Shows a default option."""
    config = context.obj["config"]
    if option in config:
        echov(f"The value of '{option}' is set to {config[option]}.")
    else:
        echow(f"The value of '{option}' is not set!")


@config.command()
@click.pass_context
def clear(context):
    """ Unsets a default option."""
    config_path = context.obj["config_path"]

    if click.confirm(f"Do you really want to clear the configuration file?"):
        # Erase content of configuration file
        write_config({}, config_path)
        echov("Configuration file cleared!")
    else:
        echov("Aborted! Nothing changed.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ytscraper', 'ytscraper.commands', 'ytscraper.helper']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.3,<2.0.0',
 'click>=7.1.1,<8.0.0',
 'google-api-python-client>=1.8.0,<2.0.0',
 'ratelimit>=2.2.1,<3.0.0',
 'toml>=0.10.0,<0.11.0']

entry_points = \
{'console_scripts': ['yt-scraper = ytscraper.app:run']}

setup_kwargs = {
    'name': 'yt-scraper',
    'version': '0.6.0',
    'description': 'Command line utility querying the YouTube API v3.',
    'long_description': '# YouTube Scraper\n> A simple command utility to extract information from the YouTube API v3 for scientific purposes.\n\n[![made-with-python](https://img.shields.io/badge/Made%20with-Python-1f425f.svg)](https://www.python.org/)\n[![License: Unlicense](https://img.shields.io/badge/license-Unlicense-blue.svg)](http://unlicense.org/)\n[![version](https://img.shields.io/pypi/v/yt-scraper)](https://pypi.org/project/yt-scraper/)\n\n## About\n\nThis Python based command line utility enables the easy extraction of information from the YouTube API (Version 3). Currently, it only supports a small subset of functions of the API interface and focuses on extracting related videos from a given starting point.\n\n\n## Installation\n\nInstall **yt-scraper** by using [pip][pip-url]:\n```sh\nsudo -H pip install yt-scraper\n```\n\nUpdate by adding the `--upgrade` flag:\n\n```sh\nsudo -H pip install --upgrade yt-scraper\n```\n\n## Usage\n### Commands\nCurrently, there is only one command supported by **yt-scraper**: *search*\n\n#### search\nThe search command starts a video search from a given starting point, such as a search term or a video itself.\n\nFor example the following command will return the first video when one searches for `cat`.\n\n```sh\n$ yt-scraper search term \'cat\'\n```\n> VideoNode(videoId=\'0A2R27kCeD4\', depth=0, rank=0, relatedVideos=(\'XewbmK0kmpI\',))\n\nOne can also provide a video id or a video url as a starting point, which is more interesting when used with the `--depth` option:\n\n```sh\n$ yt-scraper search id \'0A2R27kCeD4\' --depth 2\n```\n> VideoNode(videoId=\'0A2R27kCeD4\', depth=0, rank=0, relatedVideos=(\'XewbmK0kmpI\',))  \n> VideoNode(videoId=\'XewbmK0kmpI\', depth=1, rank=0, relatedVideos=(\'hJpfROXlaPc\',))  \n> VideoNode(videoId=\'hJpfROXlaPc\', depth=2, rank=0, relatedVideos=(\'dElQqMWhDgA\',))  \n\nAdditionally, one can specify the number of videos that should be returned on each level by using the `--number` option. For example the following command returns two related videos from a given video (specified by it\'s url) and then from each sibling only one related video:\n```sh\n$ yt-scraper search url \'https://www.youtube.com/watch?v=0A2R27kCeD4\' --depth 1 --number 2 -number 1\n```\n> VideoNode(videoId=\'0A2R27kCeD4\', depth=0, rank=0, relatedVideos=(\'XewbmK0kmpI\', \'U5KLMeFK_UY\'))  \n> VideoNode(videoId=\'XewbmK0kmpI\', depth=1, rank=0, relatedVideos=(\'hJpfROXlaPc\',))  \n> VideoNode(videoId=\'U5KLMeFK_UY\', depth=1, rank=1, relatedVideos=(\'nFrb-C6I6Ps\',))  \n\nFor the sake of brevity, you can shorten `--number` to `-n` and `--depth` to `-d`.\n\n##### Options\n\n| Search options    | Default    | Description                                                                             |\n|-------------------|------------|-----------------------------------------------------------------------------------------|\n| `-n`, `--number`  | 1          | Number of the videos fetched per level. Can be specified multiple times for each level. |\n| `-d`, `--max-depth`   | 0          | Number of recursion steps to perform.                                                   |\n| `-k`, `--api-key` | *Required* | The API key that should be used to query the YouTube API v3.                            |\n\n### Global Options\nGlobal options are specified before the command. For example, to get more output during the program execution, specify `-v` right after `yt-scraper`:\n\n```sh\n$ yt-scraper -v search term \'cat\'\n```\n\nAll global options:\n| Global options        | Default           | Description                                                                                                       |\n|-----------------------|-------------------|-------------------------------------------------------------------------------------------------------------------|\n| `-c`, `--config-path` | *System-specific* | Specifies a configuration file. For details, see [configuration](#Configuration). |\n| `-v`, `--verbose`     | False             | Shows more output during program execution.                                                                       |\n\n\n## Configuration\nInstead of repeatedly passing the same options to yt-scraper, one can specify these options in a `config.toml` file. These values will be used in all future queries as long as they are not get overwritten by actual command line options.\n\nFor example, to always use the API key `ABCDEFGH` and a search depth of 3, where on each level one video less is returned, just create following configuration file:\n\n**config.toml**\n```toml\napi_key = "ABCDEFGH"\nnumber = [ 4, 3, 2, 1 ]\ndepth = 3\nverbose = true\n```\nA example toml is included: [config.toml][config-url]\n\nThen put this file in your standard configuration folder. Typically this folder can be found at the following system-specific locations:\n\n- Mac OS X: `~/Library/Application Support/YouTube Scraper`\n- Unix: `~/.config/youtube-scraper`\n- Windows: `C:\\Users\\<user>\\AppData\\Roaming\\YouTube Scraper`\n\nIf the folder does not exist, you may need to create it.\n\n\n## Release History\n\n* 0.2.6 \n    - Added [UNLICENSE](license-url) to project.\n* 0.3.0\n    - Uploaded to [PyPI][pypi-url]\n* 0.4.0\n    - New command *search*\n* 0.5.0\n    - Option `--depth` renamed to `--max-depth`\n    - Video attributes, such as title, description, channel are fetched.\n    - More consistent option handling\n* 0.6.0\n    - New export feature: *csv*\n    - New command: *config*\n    - New API options: *region-code, lang-code and safe-search*\n\n\n## Roadmap\n\nEvery of these features is going to be a minor patch:\n\n- [X] Add node video data attributes, such as title and description.\n- [ ] Add possibility to specify more than one API key to switch seamlessly.\n- [ ] Add possibility to query more than 50 videos on one level.\n- [ ] Add youtube-dl integration for downloading subtitles.\n- [ ] Add a testing suite.\n- [o] Add export functionality to CSV, SQLlite or Pandas.\n \n\n## Contributing\nIf you found a bug or have a suggestion, \nplease don\'t hesitate to [file an issue][git-new-issue-url].\n\nContributions in any form are welcomed. \nI will accept pull-requests if they extent **yt-scraper**\'s functionality.\n\nTo set up the development environment, \nplease install [Poetry][poetry-url] and run `poetry install` inside the project.\nA test suite will be added soon.\n\nIn general, the contribution process is somehow like this:\n\n1. Fork it (`$ git clone https://github.com/rattletat/yt-scraper`)\n2. Create your feature branch (`$ git checkout -b feature/fooBar`)\n3. Commit your changes (`$ git commit -am \'Add some fooBar\'`)\n4. Push to the branch (`$ git push origin feature/fooBar`)\n5. Create a new Pull Request\n\n\n## Author\n**Michael Brauweiler**\n\n- Twitter: [@rattletat][me-twitter-url]\n- Email: [rattletat@posteo.me](mailto:rattletat@posteo.me)\n\n\n## License\nThis plugin is free and unemcumbered software released into the public domain. \n\nFor more information, see the included [UNLICENSE][license-url] file.\n\n<!-- Markdown link & img dfn\'s -->\n[pip-url]: https://pip.pypa.io/en/stable/\n[config-url]: data/config.toml\n[git-new-issue-url]: https://github.com/rattletat/yt-scraper/issues/new\n[poetry-url]: https://github.com/python-poetry/poetry\n[pypi-url]: https://pypi.org/project/yt-scraper/\n[me-github-url]: https://github.com/rattletat\n[me-twitter-url]: https://twitter.com/m_brauweiler\n[license-url]: UNLICENSE\n',
    'author': 'Michael Brauweiler',
    'author_email': 'rattletat@posteo.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
import csv
import os


def export_to_csv(nodes, output_dir):
    nodes_path = os.path.join(output_dir, "nodes.csv")
    edges_path = os.path.join(output_dir, "edges.csv")

    edges = set(
        (node["videoId"], child) for node in nodes for child in node["relatedVideos"]
    )
    print(edges)
    with open(edges_path, "w", newline="") as out:
        csv_out = csv.writer(out)
        csv_out.writerow(["origin", "target"])
        for origin, target in edges:
            csv_out.writerow([origin, target])

    with open(nodes_path, "w", newline="") as out:
        csv_out = csv.writer(out)
        csv_out.writerow(nodes[0].keys())
        for node in nodes:
            csv_out.writerow(node.values())
